import asyncio
import logging
from datetime import datetime
from typing import Any, Callable, List, Tuple, Union

import pyarrow as pa
from pydantic import BaseModel
from pypika import MySQLQuery as PikaQuery
from pypika import Order
from pypika import Table as PikaTable
from tabulate import tabulate

from ...const import DT_FMT, TZ
from ...model import BaseActivity, BaseToken, Column, Metadata, SelectOption, Table, UserInfo, View
from ...model.column import COLUMN_DATA
from ...serde import Deserializer, FromPython, ToPython
from ...utils import divide_chunks, parse_str_datetime
from ..conf import SEATABLE_URL
from ..core import TABULATE_CONF
from .builtin import BuiltInBaseClient

logger = logging.getLogger()

FIRST_COLUMN_TYPES = ["text", "number", "date", "single-select", "formular", "autonumber"]


class LinkValueNotExists(Exception):
    pass


class ColumnNotExists(Exception):
    pass


class KeyNotUnique(Exception):
    pass


################################################################
# BaseClient
################################################################
class BaseClient(BuiltInBaseClient):
    ################################################################
    # List/Get Table, View, Column
    ################################################################
    # List Tables
    async def list_tables(self, refresh: bool = True):
        metadata = await self.get_metadata(refresh=refresh)
        tables = metadata.tables
        return tables

    # Get Table
    async def get_table(self, table_name: str, refresh: bool = True):
        tables = await self.list_tables(refresh=refresh)
        for table in tables:
            if table.name == table_name:
                return table
        else:
            raise KeyError()

    # Get Table by ID
    async def get_table_by_id(self, table_id: str, refresh: bool = True):
        tables = await self.list_tables(refresh=refresh)
        for table in tables:
            if table.id == table_id:
                return table
        else:
            raise KeyError()

    # List Views
    # [NOTE] BuiltIn에 있음

    # Get View
    # [NOTE] BuiltIn에 있음

    # Get View by ID
    async def get_view_by_id(self, table_id: str, view_id: str, refresh: bool = True):
        table = await self.get_table_by_id(table_id=table_id, refresh=refresh)
        for view in table.views:
            if view.id == view_id:
                break
        else:
            _msg = f"no view id '{view_id}' in table '{table.name}'!"
            raise KeyError(_msg)

        view = await self.get_view(table_name=table.name, view_name=view.name)

        return view

    # List Columns
    async def list_columns(self, table_name: str, refresh: bool = True):
        table = await self.get_table(table_name=table_name, refresh=refresh)
        return table.columns

    # Get Column
    async def get_column(self, table_name: str, column_name: str, refresh: bool = True):
        table = await self.get_table(table_name=table_name, refresh=refresh)
        for column in table.columns:
            if column.name == column_name:
                return column
        else:
            _msg = f"no column (name: {column_name}) in table (name: {table_name})."
            raise KeyError(_msg)

    # Get Column by ID
    async def get_column_by_id(self, table_id: str, column_id: str, refresh: bool = True):
        table = await self.get_table_by_id(table_id=table_id, refresh=refresh)
        for column in table.columns:
            if column.key == column_id:
                return column
        else:
            _msg = f"no column (id: {column_id}) in table (id: {table_id})."
            raise KeyError(_msg)

    # Get 1st Column
    async def get_first_column(self, table_name: str, refresh: bool = True):
        table = await self.get_table(table_name=table_name, refresh=refresh)
        for column in table.columns:
            if column.key == "0000":
                return column
        else:
            _msg = f"column key '0000' not found!"
            raise KeyError(_msg)

    # ls
    async def ls(self, table_name: str = None):
        metadata = await self.get_metadata()
        tables = metadata.tables
        if table_name:
            for table in tables:
                if table.name == table_name:
                    break
            else:
                raise KeyError()
            columns = [{"key": c.key, "name": c.name, "type": c.type} for c in table.columns]
            print(tabulate(columns, **TABULATE_CONF))
            return
        _tables = list()
        for table in tables:
            _n = len(table.columns)
            _columns = ", ".join(c.name for c in table.columns)
            if len(_columns) > 50:
                _columns = _columns[:50] + "..."
            _columns += f" ({_n})"
            _tables += [
                {
                    "id": table.id,
                    "name": table.name,
                    "views": ", ".join([v.name for v in table.views]),
                    "columns": _columns,
                },
            ]
        print(tabulate(_tables, **TABULATE_CONF))

    ################################################################
    # ROWS
    ################################################################
    # (OVERRIDE) Append Rows
    # [NOTE] Rate Limit 조정하는 로직 필요, 지금은 금방 Rate Limit 걸릴 것 같음.
    async def append_rows(
        self, table_name: str, rows: List[dict], add_link_if_not_exists: bool = False, refresh: bool = True
    ):
        link_columns = await self.list_link_columns(table_name=table_name, refresh=refresh)
        link_column_names = [x.name for x in link_columns]

        list_links = list()
        for row in rows:
            links = dict()
            for lc in link_column_names:
                if lc in row:
                    value = row.pop(lc)
                    links.update({lc: value})
            list_links.append(links)

        # link 없으면 그냥 append_rows 수행 (성능 개선 및 Rate Limit 절약)
        if not any(list_links):
            return await super().append_rows(table_name=table_name, rows=rows)

        # 순서보장 위해서 gather 사용하지 않음
        add_rows_results = list()
        for row in rows:
            add_row_result = await self.add_row(table_name=table_name, row=row)
            add_rows_results.append(add_row_result)

        # gather 사용하여도 순서보장
        # [NOTE] create_row_links_batch로 수정하면 Rate Limit 조금 더 아낄 수 있을 것
        coros_create_row_links = list()
        for result, links in zip(add_rows_results, list_links):
            if not links:
                continue
            for column_name, display_values in links.items():
                kwargs = await self._prep_create_row_links(
                    table_name=table_name,
                    column_name=column_name,
                    display_values=display_values,
                    add_link_if_not_exists=add_link_if_not_exists,
                )
                coros_create_row_links.append(self.create_row_links(row_id=result["_id"], **kwargs))
        create_links_results = await asyncio.gather(*coros_create_row_links)
        for result in create_links_results:
            if isinstance(result, Exception):
                raise result

        return {"inserted_rows": len(add_rows_results)}

    # (OVERRIDE) Update Rows
    async def update_rows(
        self, table_name: str, updates: List[dict], add_link_if_not_exists: bool = False, refresh: bool = True
    ):
        link_columns = await self.list_link_columns(table_name=table_name, refresh=refresh)
        link_column_names = [x.name for x in link_columns]

        # 순서에 따라 동작! asyncio.gather 금지!
        coros_create_row_links = list()
        for up in updates:
            for column_name in link_column_names:
                if column_name in up["row"]:
                    display_values = up["row"].pop(column_name)
                    kwargs = await self._prep_create_row_links(
                        table_name=table_name,
                        column_name=column_name,
                        display_values=display_values,
                        add_link_if_not_exists=add_link_if_not_exists,
                    )
                    coros_create_row_links.append(self.create_row_links(row_id=up["row_id"], **kwargs))

        # update rows
        update_rows_results = await super().update_rows(table_name=table_name, updates=updates)

        # create row links
        create_links_results = await asyncio.gather(*coros_create_row_links)
        for result in create_links_results:
            if isinstance(result, Exception):
                raise result
            if not result["success"]:
                _msg = "create_row_links failed!"
                logger.error(_msg)

        return update_rows_results

    # Validate Input Columns
    async def _validate_input_columns(self, table_name: str, rows: List[dict], refresh: bool = True):
        columns = await self.list_columns(table_name=table_name, refresh=refresh)
        column_names = [c.name for c in columns]

        input_columns = set()
        _ = [input_columns.add(k) for row in rows for k in row]

        for input_column in input_columns:
            if input_column not in column_names:
                _msg = f"table {table_name} does not have column '{input_column}'!"
                raise ColumnNotExists(_msg)

    # Upsert Rows - 궁극의 메쏘드!
    # [NOTE] 첫 Column을 Unique하게 사용하기만 한다면, 이 메쏘드 하나만 써서 Append, Update 해결 가능!
    async def upsert_rows(
        self,
        table_name: str,
        rows: List[dict],
        key_column: str = None,
        add_link_if_not_exists: bool = False,
        raise_key_not_unique_error: bool = True,
    ):
        # correct input
        rows = rows if isinstance(rows, list) else [rows]

        # validate
        await self._validate_input_columns(table_name=table_name, rows=rows, refresh=True)

        # default key column is first column
        if not key_column:
            first_column = await self.get_first_column(table_name=table_name, refresh=False)
            key_column = first_column.name

        # get row id map
        id_map = await self.get_row_id_map(
            table_name=table_name, key_column=key_column, raise_key_not_unique_error=raise_key_not_unique_error
        )

        # split updates & appends
        updates, appends = list(), list()
        for row in rows:
            key = row.get(key_column)
            if key is None or key not in id_map:
                appends.append(row)
            else:
                updates.append({"row_id": id_map[key], "row": row})

        # 1. Key Column 값 존재하면 Update
        update_coro = self.update_rows(
            table_name=table_name, updates=updates, add_link_if_not_exists=add_link_if_not_exists
        )

        # 2. Key Column 값 존재하지 않으면 Append
        append_coro = self.append_rows(
            table_name=table_name, rows=appends, add_link_if_not_exists=add_link_if_not_exists
        )

        update_results, append_results = await asyncio.gather(update_coro, append_coro)

        return {**update_results, **append_results}

    ################################################################
    # QUERY
    ################################################################
    # Get Row ID Map
    async def get_row_id_map(
        self, table_name: str, key_column: str = None, raise_key_not_unique_error: bool = True, refresh: bool = True
    ):
        if not refresh and table_name in self.row_id_map and key_column in self.row_id_map[table_name]:
            return self.row_id_map[table_name][key_column]

        if key_column is None:
            first_column = await self.get_first_column(table_name=table_name)
            key_column = first_column.name
        rows = await self.read_table(table_name=table_name, select=["_id", key_column])

        # check key column is unique
        l = [r[key_column] for r in rows]
        s = set(l)
        if len(l) != len(s):
            _dups = list()
            for e in s:
                cnt = l.count(e)
                if cnt > 1:
                    _dups.append(f"{e} * {cnt}")
            _dups = ", ".join(_dups)
            _msg = f"key column is not unique. do not use upsert method! (dups: {_dups})"
            if raise_key_not_unique_error:
                raise KeyNotUnique(_msg)
            logger.warning(_msg)

        if table_name not in self.row_id_map:
            self.row_id_map[table_name] = dict()
        self.row_id_map[table_name].update({key_column: {r[key_column]: r["_id"] for r in rows}})

        return self.row_id_map[table_name][key_column]

    async def _read_table(
        self,
        table_name: str,
        select: List[str] = None,
        modified_before: str = None,
        modified_after: str = None,
        order_by: str = None,
        desc: bool = False,
        offset: int = 0,
        limit: int = None,
    ):
        MAX_LIMIT = 10000
        OFFSET = 0

        # correct args
        table = PikaTable(table_name)
        if not select:
            select = ["*"]
        if not isinstance(select, list):
            select = [x.strip() for x in select.split(",")]
        _limit = min(MAX_LIMIT, limit) if limit else limit
        _offset = offset if offset else OFFSET

        # generate query
        q = PikaQuery.from_(table).select(*select)

        if modified_before or modified_after:
            last_modified = "_mtime"
            tbl = await self.get_table(table_name=table_name)
            for c in tbl.columns:
                if c.key == "_mtime":
                    last_modified = c.name
                    break
            if modified_after:
                if isinstance(modified_after, datetime):
                    modified_after = modified_after.isoformat(timespec="milliseconds")
                q = q.where(table[last_modified] > modified_after)
            if modified_before:
                if isinstance(modified_before, datetime):
                    modified_before = modified_before.isoformat(timespec="milliseconds")
                q = q.where(table[last_modified] < modified_before)

        if order_by:
            q = q.orderby(order_by, order=Order.desc if desc else Order.asc)

        q = q.limit(_limit or MAX_LIMIT)

        # 1st hit
        rows = await self.list_rows_with_sql(sql=q.offset(_offset))

        # get all records
        if not limit or len(rows) < limit:
            while True:
                _offset += MAX_LIMIT
                _rows = await self.list_rows_with_sql(sql=q.offset(_offset))
                rows += _rows
                if len(_rows) < MAX_LIMIT:
                    break

        return rows

    # read table with schema
    async def read_table_with_schema(
        self,
        table_name: str,
        select: List[str] = None,
        modified_before: str = None,
        modified_after: str = None,
        order_by: str = None,
        desc: bool = False,
        offset: int = 0,
        limit: int = None,
        Deserializer: Deserializer = ToPython,
    ) -> dict:
        # list rows
        rows = await self._read_table(
            table_name=table_name,
            select=select,
            modified_before=modified_before,
            modified_after=modified_after,
            order_by=order_by,
            desc=desc,
            offset=offset,
            limit=limit,
        )

        if not Deserializer:
            _msg = "Deserializer required!"
            raise KeyError(_msg)

        # to python data type
        metadata = await self.get_metadata()
        collaborators = await self.list_collaborators()
        deserializer = Deserializer(
            metadata=metadata,
            table_name=table_name,
            base_name=self.base_name,
            group_name=self.group_name,
            collaborators=collaborators,
        )
        try:
            rows = deserializer(*rows, select=select)
        except Exception as ex:
            _msg = f"deserializer failed - group '{self.group_name}', base '{self.base_name}', table '{table_name}'"
            logger.error(_msg)
            raise ex

        return {
            "table": deserializer.schema(),
            "rows": rows,
            "last_mtime": deserializer.last_modified,
        }

    # read table
    async def read_table(
        self,
        table_name: str,
        select: List[str] = None,
        modified_before: str = None,
        modified_after: str = None,
        order_by: str = None,
        desc: bool = False,
        offset: int = 0,
        limit: int = None,
        Deserializer: Deserializer = ToPython,
    ) -> List[dict]:
        # list rows
        rows = await self._read_table(
            table_name=table_name,
            select=select,
            modified_before=modified_before,
            modified_after=modified_after,
            order_by=order_by,
            desc=desc,
            offset=offset,
            limit=limit,
        )

        # deserializer
        if Deserializer:
            metadata = await self.get_metadata()
            collaborators = await self.list_collaborators()
            deserializer = Deserializer(
                metadata=metadata,
                table_name=table_name,
                base_name=self.base_name,
                group_name=self.group_name,
                collaborators=collaborators,
            )
            try:
                rows = deserializer(*rows, select=select)
            except Exception as ex:
                _msg = (
                    f"deserializer failed - group '{self.group_name}', base '{self.base_name}', table '{table_name}'"
                )
                logger.error(_msg)
                raise ex

        return rows

    # read table as DataFrame
    async def read_table_as_df(
        self,
        table_name: str,
        select: List[str] = None,
        modified_before: str = None,
        modified_after: str = None,
        offset: int = 0,
        limit: int = None,
    ):
        rows = await self.read_table(
            table_name=table_name,
            select=select,
            modified_before=modified_before,
            modified_after=modified_after,
            offset=offset,
            limit=limit,
            Deserializer=ToPython,
        )

        if not rows:
            return None
        tbl = pa.Table.from_pylist(rows).to_pandas()
        return tbl.set_index("_id", drop=True).rename_axis("row_id")

    # read view
    async def read_view(
        self,
        table_name: str,
        view_name: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
        Deserializer: Deserializer = ToPython,
        return_schema: bool = False,
    ):
        rows = await self.list_rows(
            table_name=table_name,
            view_name=view_name,
            convert_link_id=convert_link_id,
            order_by=order_by,
            direction=direction,
            start=start,
            limit=limit,
        )

        # to python data type
        if Deserializer:
            metadata = await self.get_metadata()
            collaborators = await self.list_collaborators()
            deserializer = Deserializer(
                metadata=metadata,
                table_name=table_name,
                base_name=self.base_name,
                group_name=self.group_name,
                collaborators=collaborators,
            )
            try:
                rows = deserializer(*rows)
            except Exception as ex:
                _msg = f"deserializer failed - group '{self.group_name}', base '{self.base_name}', table '{table_name}', view '{view_name}'"
                logger.error(_msg)
                raise ex
            if return_schema:
                return rows, deserializer.schema()

        return rows

    # read view as DataFrame
    async def read_view_as_df(
        self,
        table_name: str,
        view_name: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
    ):
        rows = await self.read_view(
            table_name=table_name,
            view_name=view_name,
            convert_link_id=convert_link_id,
            order_by=order_by,
            direction=direction,
            start=start,
            limit=limit,
            Deserializer=ToPython,
            return_schema=False,
        )

        if not rows:
            return None
        tbl = pa.Table.from_pylist(rows).to_pandas()
        return tbl.set_index("_id", drop=True).rename_axis("row_id")

    # Generate Deserializer
    async def generate_deserializer(self, table_name: str):
        table = await self.get_table(table_name)
        users = await self.list_collaborators() if "collaborator" in [c.type for c in table.columns] else None
        return ToPython(table=table, users=users)

    # get last modified at
    async def get_last_mtime(self, table_name: str):
        table = await self.get_table(table_name=table_name)
        for column in table.columns:
            if column.type == "mtime":
                c = column.name
                q = f"SELECT {c} FROM {table_name} ORDER BY {c} DESC LIMIT 1;"
                r = await self.list_rows_with_sql(q)
                last_mtime = parse_str_datetime(r[0][c])
                return last_mtime
        else:
            raise KeyError

    ################################################################
    # LINKS
    ################################################################
    # List Link Columns
    async def list_link_columns(self, table_name: str, refresh: bool = True):
        table = await self.get_table(table_name=table_name, refresh=refresh)

        link_columns = list()
        for c in table.columns:
            if c.type != "link":
                continue
            link_columns.append(c)

        return link_columns

    # Prep - Create Row Link
    async def _prep_create_row_link(
        self,
        table_name: str,
        column_name: str,
        display_value: Union[str, int, float, datetime],
        add_link_if_not_exists: bool = False,
    ):
        prep = await self._prep_create_row_links(
            table_name=table_name,
            column_name=column_name,
            display_values=[display_value],
            add_link_if_not_exists=add_link_if_not_exists,
        )

        other_rows_ids = prep.pop("other_rows_ids")
        prep.update({"other_table_row_id": other_rows_ids[0]})
        return prep

    # Prep - Create Row Links
    async def _prep_create_row_links(
        self,
        table_name: str,
        column_name: str,
        display_values: list,
        add_link_if_not_exists: bool = False,
        raise_key_not_unique_error: bool = True,
    ):
        # correct display values
        display_values = display_values if isinstance(display_values, list) else [display_values]

        table = await self.get_table(table_name=table_name, refresh=True)
        column = await self.get_column(table_name=table_name, column_name=column_name, refresh=False)
        if column.type != "link":
            _msg = f"type of column '{column_name}' is not link type."
            raise KeyError(_msg)

        is_host = True if table.id == column.data["table_id"] else None

        if is_host:
            _table = table
            _other_table = await self.get_table_by_id(table_id=column.data["other_table_id"], refresh=False)
        else:
            _table = await self.get_table_by_id(table_id=column.data["table_id"], refresh=False)
            _other_table = table
        display_column_key = column.data["display_column_key"]

        display_column = await self.get_column_by_id(
            table_id=_other_table.id, column_id=display_column_key, refresh=False
        )
        row_id_map = await self.get_row_id_map(
            table_name=_other_table.name,
            key_column=display_column.name,
            raise_key_not_unique_error=raise_key_not_unique_error,
        )

        other_rows_ids = list()
        for display_value in display_values:
            if display_value not in row_id_map:
                if not add_link_if_not_exists:
                    _msg = f"display value '{display_value}' not exists. please add this value into table '{_other_table.name}' first."
                    raise LinkValueNotExists(_msg)
                # add link
                await self.add_row(table_name=_other_table.name, row={display_column.name: display_value})
                row_id_map = await self.get_row_id_map(
                    table_name=_other_table.name,
                    key_column=display_column.name,
                    raise_key_not_unique_error=raise_key_not_unique_error,
                )
            other_rows_ids.append(row_id_map[display_value])

        return {
            "table_name": _table.name,
            "other_table_name": _other_table.name,
            "link_id": column.data["link_id"],
            "other_rows_ids": other_rows_ids,
        }

    # Get Ohter Rows Ids
    async def get_other_rows_ids(self, table_name, column_name, raise_key_not_unique_error: bool = True):
        link = await self.get_link(table_name=table_name, column_name=column_name)
        other_table = await self.get_table_by_id(table_id=link["other_table_id"])
        for column in other_table.columns:
            if column.key == link["display_column_key"]:
                break
        else:
            raise KeyError
        return await self.get_row_id_map(
            table_name=other_table.name, key_column=column.name, raise_key_not_unique_error=raise_key_not_unique_error
        )

    ################################################################
    # FILES & IMAGES
    ################################################################

    ################################################################
    # TABLES
    ################################################################

    # Create New Table
    # [NOTE] 이 endpoint는 Link 컬럼을 처리하지 못 함. (2023.9.9 현재)
    async def _create_new_table(self, table_name: str, columns: List[dict] = None):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {"table_name": table_name}
        if columns:
            json.update({"columns": columns})

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Create Table
    # [NOTE] 현재 Create New Table API 문제 때문에 사용 - 2번째 Colmnn부터는 insert_column으로 추가.
    async def create_table(self, table_name: str, columns: List[dict], overwrite: bool = False):
        tables = await self.list_tables()
        if table_name in [t.name for t in tables]:
            if not overwrite:
                _msg = f"table '{table_name}' already exists!"
                raise KeyError(_msg)
            r = await self.delete_table(table_name=table_name)
            if not r["success"]:
                _msg = f"delete table '{table_name}' failed!"
                raise KeyError(_msg)

        # seprate key column
        key_column, columns = columns[0], columns[1:]
        if key_column["column_type"] not in FIRST_COLUMN_TYPES:
            _msg = f"""only '{", ".join(FIRST_COLUMN_TYPES)}' can be a first column"""
            raise KeyError(_msg)

        # create table
        _ = await self._create_new_table(table_name=table_name, columns=[key_column])

        # insert columns
        for column in columns:
            _ = await self.insert_column(table_name=table_name, **column)

        return True

    ################################################################
    # IMPORT
    ################################################################
    # (DEPRECATED) Create Table from CSV

    # (DEPRECATED) Append Rows from CSV

    ################################################################
    # VIEWS
    ################################################################
    # Get View by ID

    ################################################################
    # COLUMNS
    ################################################################

    # add select options if not exists
    async def add_select_options_if_not_exists(self, table_name: str, rows: List[dict]):
        table = await self.get_table(table_name=table_name)
        columns_and_options = {
            c.name: [o["name"] for o in c.data["options"]] if c.data else []
            for c in table.columns
            if c.type in ["single-select", "multiple-select"]
        }

        if not columns_and_options:
            return

        options = {c: set([r.get(c) for r in rows if r.get(c)]) for c in columns_and_options}
        options_to_add = dict()
        for column_name, column_options in options.items():
            for column_opt in column_options:
                if column_opt not in columns_and_options[column_name]:
                    if column_name not in options_to_add:
                        options_to_add[column_name] = list()
                    options_to_add[column_name].append(SelectOption(name=column_opt))

        coros = [
            self.add_select_options(table_name=table_name, column_name=column_name, options=options)
            for column_name, options in options_to_add.items()
        ]

        return await asyncio.gather(*coros)

    ################################################################
    # BIG DATA
    ################################################################

    ################################################################
    # ROW COMMENTS
    ################################################################

    ################################################################
    # NOTIFICATION
    ################################################################

    ################################################################
    # ACTIVITIES & LOGS
    ################################################################
    # List Delete Operation Logs After
    async def list_delete_operation_logs_since(self, op_type: str, op_time: Union[datetime, str], per_page: int = 100):
        # correct op_time
        op_time = datetime.fromisoformat(op_time) if isinstance(op_time, str) else op_time

        delete_logs = list()
        page = 1
        while True:
            logs = await self.list_delete_operation_logs(op_type=op_type, page=page, per_page=per_page)
            if not logs:
                break
            for log in logs:
                if datetime.fromisoformat(log["op_time"]) < op_time:
                    break
                delete_logs.append(log)
            page += 1

        return delete_logs

    ################################################################
    # SNAPSHOTS
    ################################################################
    # TBD



import logging
from datetime import date, datetime
from typing import Any, List, Union

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import ARRAY, BOOLEAN, DATE, FLOAT, INTEGER, SMALLINT, TEXT, TIMESTAMP, VARCHAR

from plantable.const import DT_FMT, TZ

from ...model import Column, Table, User, Metadata
from .deserializer import ColumnDeserializer, Deserializer

logger = logging.getLogger(__name__)


################################################################
# Postgres Types for SeaTable
################################################################
class PostgresId(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, VARCHAR(255), nullable=False, primary_key=True)

    def convert(self, x):
        return str(x)


class PostgresBool(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, BOOLEAN, nullable=True)

    def convert(self, x):
        return bool(x)


class PostgresText(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, TEXT, nullable=True)

    def convert(self, x):
        return str(x)


class PostgresEmail(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, VARCHAR(2083), nullable=True)

    def convert(self, x):
        return str(x)


class PostgresUrl(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, VARCHAR(2083), nullable=True)

    def convert(self, x):
        return str(x)


class PostgresRate(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, SMALLINT, nullable=True)

    def convert(self, x):
        return int(x)


class _PostgresInteger(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, INTEGER, nullable=True)

    def convert(self, x):
        return int(x)


class _PostgresFloat(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, FLOAT, nullable=True)

    def convert(self, x):
        return float(x)


class PostgresNumber(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )
        if self.data.get("enable_precision"):
            self.sub_deserializer = _PostgresInteger(name=self.name, seatable_type=self.seatable_type, data=self.data)
        else:
            self.sub_deserializer = _PostgresFloat(name=self.name, seatable_type=self.seatable_type, data=self.data)

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        return self.sub_deserializer(x)


class _PostgresDate(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, DATE, nullable=True)

    def convert(self, x):
        return date.fromisoformat(x[:10])


class _PostgresDatetime(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, TIMESTAMP(timezone=True), nullable=True)

    def convert(self, x):
        if x.endswith("Z"):
            x = x.replace("Z", "+00:00", 1)
        try:
            x = datetime.strptime(x, DT_FMT)
        except Exception as ex:
            x = datetime.fromisoformat(x)
        return x.astimezone(TZ)


class PostgresDate(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )
        if self.data and self.data["format"] == "YYYY-MM-DD":
            self.sub_deserializer = _PostgresDate(name=self.name, seatable_type=self.seatable_type, data=self.data)
        else:
            self.sub_deserializer = _PostgresDatetime(name=self.name, seatable_type=self.seatable_type, data=self.data)

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        return self.sub_deserializer(x)


class PostgresDuration(ColumnDeserializer):
    def schema(self):
        # [TODO] 지금은 "초"만 사용하는 정수 - 나중에 바꿀 것
        return sa.Column(self.name, INTEGER, nullable=True)

    def convert(self, x):
        return int(x)


class PostgresSingleSelect(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, VARCHAR(255), nullable=True)

    def convert(self, x):
        return str(x)


class PostgresMultipleSelect(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, ARRAY(VARCHAR(255)), nullable=True)

    def convert(self, x):
        return [str(_x) for _x in x]


class PostgresUser(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, VARCHAR(255), nullable=True)

    def convert(self, x):
        if not self.collaborator_map:
            return x
        return self.collaborator_map[x] if x in self.collaborator_map else x


class PostgresListcollaborator_map(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, ARRAY(VARCHAR(255)), nullable=True)

    def convert(self, x):
        if not self.collaborator_map:
            return x
        return [self.collaborator_map[_x] if _x in self.collaborator_map else _x for _x in x]


class PostgresFile(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, ARRAY(VARCHAR(2083)), nullable=True)

    def convert(self, x):
        return [_x["url"] for _x in x]


class PostgresImage(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, ARRAY(VARCHAR(2083)), nullable=True)

    def convert(self, x):
        return x


class PostgresAutoNumber(ColumnDeserializer):
    def schema(self):
        return sa.Column(self.name, VARCHAR(255), nullable=True)

    def convert(self, x):
        return str(x)


DESERIALIZER = {
    "row-id": PostgresId,
    "checkbox": PostgresBool,
    "bool": PostgresBool,  # formula가 result_type으로 bool을 사용
    "text": PostgresText,
    "string": PostgresText,  # formula가 result_type으로 string을 사용
    "button": PostgresText,
    "long-text": PostgresText,
    "email": PostgresEmail,
    "url": PostgresUrl,
    "rate": PostgresRate,
    "number": PostgresNumber,
    "date": PostgresDate,
    "duration": PostgresDuration,
    "ctime": PostgresDate,
    "mtime": PostgresDate,
    "single-select": PostgresSingleSelect,
    "multiple-select": PostgresText,
    "user": PostgresUser,
    "collaborator": PostgresListcollaborator_map,
    "creator": PostgresUser,
    "last-modifier": PostgresUser,
    "file": PostgresFile,
    "image": PostgresImage,
    "auto-number": PostgresAutoNumber,
}


class PostgresFormula(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )
        self.sub_deserializer = DESERIALIZER[self.data["result_type"]](
            name=self.name, seatable_type=self.data["result_type"], data=dict()
        )

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        if x == "#VALUE!":
            return None
        return self.sub_deserializer(x)


class PostgresLink(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )

        self.sub_deserializer = DESERIALIZER[self.data["array_type"]](
            name=self.name, seatable_type=self.data["array_type"], data=self.data["array_data"]
        )
        self.is_multiple = self.data["is_multiple"]

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        x = [self.sub_deserializer(_x.get("display_value")) for _x in x]
        if not x:
            return None
        if self.is_multiple:
            return x
        return x[0]


class PostgresLinkFormula(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )

        self.sub_deserializer = DESERIALIZER[self.data["array_type"]](
            name=self.name, seatable_type=self.data["array_type"], data=self.data["array_data"]
        )

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        return [self.sub_deserializer(_x) for _x in x]


DESERIALIZER.update({"formula": PostgresFormula, "link": PostgresLink, "link-formula": PostgresText})


################################################################
# Seatable to Postgres
################################################################
class ToPostgres(Deserializer):
    Deserializer = DESERIALIZER

    def schema(self):
        name = self.generate_table_name()
        columns = [column.schema() for _, column in self.columns.items()]
        return sa.Table(name, sa.MetaData(), *columns)



import logging
from datetime import date, datetime
from typing import Any, List, Union

import requests

from ...const import DT_FMT, SYSTEM_FIELDS, TZ
from ...model import Table

logger = logging.getLogger(__name__)


################################################################
# Converter
################################################################
class FromPython:
    def __init__(self, table: Table, overwrite_none: bool = False):
        self.table_name = table.name
        self.overwrite_none = overwrite_none

        self.schema = {column.name: column for column in table.columns}

    def __call__(self, row):
        if row is None:
            return
        row = {
            column: getattr(self, self.schema[column].type.replace("-", "_"))(
                value=value, data=self.schema[column].data
            )
            for column, value in row.items()
            if column in self.schema
        }
        row = {k: v for k, v in row.items() if v != "__ignore__"}
        if not self.overwrite_none:
            row = {k: v for k, v in row.items() if v is not None}
        return row

    @staticmethod
    def _ensure_list(x):
        if x is None:
            return x
        return x if isinstance(x, list) else [x]

    def checkbox(self, value, data: dict = None) -> bool:
        return value

    def text(self, value: str, data: dict = None) -> str:
        return value

    def button(self, value: str, data: dict = None) -> str:
        return value

    def long_text(self, value: str, data: dict = None) -> str:
        return value

    def email(self, value: str, data: dict = None) -> str:
        return value

    def url(self, value: str, data: dict = None) -> str:
        return value

    def rate(self, value: int, data: dict = None) -> int:
        return value

    def number(self, value: Union[int, float], data: dict = None) -> Union[int, float]:
        return value

    def date(self, value: Union[date, datetime], data: dict = None) -> Union[date, datetime]:
        return str(value)

    def duration(self, value: Union[str, int], data: dict = None) -> int:
        # [TODO] currently H:MM:SS only
        if isinstance(value, int):
            h, value = value // 3600, value % 3600
            m, s = value // 60, value % 60
            value = f"{h}:{m:02}:{s:02}"
        return value

    def ctime(self, value: datetime, data: dict = None):
        return str(value)

    def mtime(self, value: datetime, data: dict = None):
        return str(value)

    def single_select(self, value: str, data: dict = None) -> str:
        return value

    def multiple_select(self, value: List[str], data: dict = None) -> List[str]:
        return self._ensure_list(value)

    def link(self, value: list, data: dict = None) -> list:
        # [NOTE] Link 값은 create link로 따로 입력해야 함.
        return "__ignore__"

    def link_formula(self, value, data: dict = None):
        raise KeyError("you cannot insert the value for link-formula column")

    def user(self, user: str, data: dict = None):
        return user

    def collaborator(self, value: List[str], data: dict = None) -> List[str]:
        return self._ensure_list(value)

    def creator(self, value: str, data: dict = None) -> str:
        return value

    def last_modifier(self, value: str, data: dict = None) -> str:
        return value

    def file(self, value, data: dict = None):
        return self._ensure_list(value)

    def image(self, value, data: dict = None):
        return self._ensure_list(value)

    def formula(self, value, data: dict = None):
        raise KeyError("you cannot insert the value for fomula column")

    def auto_number(self, value, data: dict = None):
        raise KeyError("you cannot insert the value for auto-number column")



import asyncio
import logging
from datetime import datetime
from typing import Any, Callable, List, Tuple, Union
import aiohttp

import pyarrow as pa
import requests
from fastapi import HTTPException, status
from pydantic import BaseModel
from pypika import MySQLQuery as PikaQuery
from pypika import Order
from pypika import Table as PikaTable
from pypika.dialects import QueryBuilder
from tabulate import tabulate

from ...const import DT_FMT, TZ
from ...model import BaseActivity, BaseToken, Column, Metadata, SelectOption, Table, UserInfo, View
from ...model.column import COLUMN_DATA
from ...serde import Deserializer, FromPython, ToPython
from ...utils import divide_chunks, parse_str_datetime
from ..conf import SEATABLE_URL
from ..core import TABULATE_CONF, HttpClient

logger = logging.getLogger()

FIRST_COLUMN_TYPES = ["text", "number", "date", "single-select", "formular", "autonumber"]


################################################################
# BuiltInBaseClient
################################################################
class BuiltInBaseClient(HttpClient):
    def __init__(
        self,
        seatable_url: str = SEATABLE_URL,
        api_token: str = None,
        base_token: BaseToken = None,
        access_token_refresh_sec: int = 86400,
    ):
        if not seatable_url:
            raise KeyError("seatable_url is required!")

        super().__init__(seatable_url=seatable_url.rstrip("/"))

        self.api_token = api_token
        self.base_token = base_token
        self.access_token_refresh_sec = access_token_refresh_sec

        if api_token:
            self.update_base_token()

        # self info
        self.dtable_uuid = self.base_token.dtable_uuid
        self.workspace_id = self.base_token.workspace_id
        self.group_id = self.base_token.group_id
        self.group_name = self.base_token.group_name
        self.base_name = self.base_token.base_name

        # store
        self.metadata = None
        self.collaborators = None
        self.views = dict()
        self.row_id_map = dict()

    # update base_token
    def update_base_token(self):
        auth_url = self.seatable_url + "/api/v2.1/dtable/app-access-token/"
        response = requests.get(auth_url, headers={"Authorization": f"Token {self.api_token}"})
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            error_msg = response.json()["error_msg"]
            if error_msg in ["Permission denied."]:
                raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Wrong base token!")
            raise ex
        results = response.json()
        self.base_token = BaseToken(**results)

    # override
    def session_maker(self):
        token_uptime = (datetime.now() - self.base_token.generated_at).seconds
        if token_uptime > self.access_token_refresh_sec:
            self.update_base_token()
            _msg = f"access token for workspace '{self.workspace_id}' is updated after {token_uptime} seconds uptime."
            logger.warning(_msg)
        headers = self.headers.copy()
        headers.update({"authorization": "Bearer {}".format(self.base_token.access_token)})
        return aiohttp.ClientSession(base_url=self.seatable_url, headers=headers)

    ################################################################
    # BASE INFO
    ################################################################
    # Get Base Info
    async def get_base_info(self):
        METHOD = "GET"
        URL = f"/dtable-server/dtables/{self.base_token.dtable_uuid}"

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        return results

    # Get Metadata
    async def get_metadata(self, model: BaseModel = Metadata, refresh: bool = True):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/metadata/"
        ITEM = "metadata"

        if not refresh and self.metadata is not None:
            return self.metadata

        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]
        if model:
            results = model(**results)

        self.metadata = results
        return self.metadata

    # Get Big Data Status
    async def get_bigdata_status(self):
        METHOD = "GET"
        URL = f"/dtable-db/api/v1/base-info/{self.base_token.dtable_uuid}/"

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        return results

    # List Collaborators
    async def list_collaborators(self, model: BaseModel = UserInfo, refresh: bool = True):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/related-users/"
        ITEM = "user_list"

        if not refresh and self.collaborators is not None:
            return self.collaborators

        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]
        if model:
            results = [model(**x) for x in results]

        self.collaborators = results
        return self.collaborators

    ################################################################
    # ROWS
    ################################################################
    # List Rows (Table, with SQL)
    async def list_rows_with_sql(self, sql: Union[str, QueryBuilder], convert_keys: bool = True):
        """
        [NOTE]
         default LIMIT 100 when not LIMIT is given!
         max LIMIT 10000!
        """
        METHOD = "POST"
        URL = f"/dtable-db/api/v1/query/{self.base_token.dtable_uuid}/"
        SUCCESS = "success"
        ITEM = "results"

        json = {"sql": sql.get_sql() if isinstance(sql, QueryBuilder) else sql, "convert_keys": convert_keys}
        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=json)
            if not response[SUCCESS]:
                raise Exception(response)
            results = response[ITEM]
        return results

    # List Rows (View)
    # [NOTE] Seatable 4.1에서 첫 Row에 없는 값은 안 읽어오는 이슈
    async def list_rows(
        self,
        table_name: str,
        view_name: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
    ):
        MAX_LIMIT = 1000

        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"
        ITEM = "rows"

        get_all_rows = False
        if not limit:
            get_all_rows = True
            limit = limit or MAX_LIMIT

        params = {
            "table_name": table_name,
            "view_name": view_name,
            "convert_link_id": str(convert_link_id).lower(),
            "order_by": order_by,
            "direction": direction,
            "start": start,
            "limit": limit,
        }

        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, **params)
            response = response[ITEM]
            results = response

            # pagination
            if get_all_rows:
                while len(response) == limit:
                    params.update({"start": params["start"] + limit})
                    response = await self.request(session=session, method=METHOD, url=URL, **params)
                    response = response[ITEM]
                    results += response

        return results

    # Add Row
    async def add_row(
        self, table_name: str, row: dict = {}, anchor_row_id: str = None, row_insert_position: str = "insert_below"
    ):
        # insert_below or insert_above
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"

        table = await self.get_table(table_name=table_name)
        serializer = FromPython(table)

        json = {"table_name": table_name, "row": serializer(row)}
        if anchor_row_id:
            json.update(
                {
                    "ahchor_row_id": anchor_row_id,
                    "row_insert_position": row_insert_position,
                }
            )

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(table_name=table_name, rows=[row])

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Update Row
    async def update_row(self, table_name: str, row_id: str, row: dict):
        # NOT WORKING
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"
        ITEM = "success"

        table = await self.get_table(table_name=table_name)
        serializer = FromPython(table=table)
        json = {"table_name": table_name, "row_id": row_id, "row": serializer(row)}

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(table_name=table_name, rows=[row])

        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=json)
            results = response[ITEM]

        return results

    # Delete Row
    async def delete_row(self, table_name: str, row_id: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"

        json = {"table_name": table_name, "row_id": row_id}

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Get Row
    async def get_row(self, table_name: str, row_id: str, convert: bool = False):
        # NOT WORKING
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/{row_id}/"

        params = {"table_name": table_name, "convert": str(convert).lower()}

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)

        return results

    # Append Rows
    async def append_rows(self, table_name: str, rows: List[dict]):
        # insert_below or insert_above
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-append-rows/"

        # get pk and serializer
        table = await self.get_table(table_name=table_name)
        serializer = FromPython(table=table)

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(table_name=table_name, rows=rows)

        # divide chunk - [NOTE] 1000 rows까지만 됨
        UPDATE_LIMIT = 1000
        chunks = divide_chunks(rows, UPDATE_LIMIT)
        list_json = [{"table_name": table_name, "rows": [serializer(r) for r in chunk]} for chunk in chunks]

        async with self.session_maker() as session:
            coros = [self.request(session=session, method=METHOD, url=URL, json=json) for json in list_json]
            list_results = await asyncio.gather(*coros)

        results = {"inserted_row_count": 0}
        for r in list_results:
            results["inserted_row_count"] += r["inserted_row_count"]

        return results

    # Update Rows
    async def update_rows(self, table_name: str, updates: List[dict]):
        # updates = [{"row_id": xxx, "row": {"key": "value"}}, ...]
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-update-rows/"

        # get serializer
        table = await self.get_table(table_name=table_name)
        serializer = FromPython(table=table)

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(
            table_name=table_name, rows=[update["row"] for update in updates]
        )

        # divide chunk - [NOTE] 1000 rows까지만 됨
        UPDATE_LIMIT = 1000
        chunks = divide_chunks(updates, UPDATE_LIMIT)
        list_json = [
            {
                "table_name": table_name,
                "updates": [{"row_id": r["row_id"], "row": serializer(r["row"])} for r in chunk],
            }
            for chunk in chunks
        ]

        async with self.session_maker() as session:
            coros = [self.request(session=session, method=METHOD, url=URL, json=json) for json in list_json]
            list_results = await asyncio.gather(*coros)

        for results in list_results:
            if isinstance(results, Exception):
                raise results

        return {"updated_row_count": len(updates)}

    # Delete Rows
    async def delete_rows(self, table_name: str, row_ids: List[str]):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-delete-rows/"

        json = {"table_name": table_name, "row_ids": row_ids}

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Rock Rows
    async def lock_rows(self, table_name: str, row_ids: List[str]):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/lock-rows/"

        json = {"table_name": table_name, "row_ids": row_ids}

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Unrock Rows
    async def unlock_rows(self, table_name: str, row_ids: List[str]):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/unlock-rows/"

        json = {"table_name": table_name, "row_ids": row_ids}

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    ################################################################
    # LINKS
    ################################################################
    # Create Row Link
    async def create_row_link(
        self, table_name: str, other_table_name: str, link_id: str, row_id: str, other_row_id: str
    ):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/links/"

        json = {
            "table_name": table_name,
            "other_table_name": other_table_name,
            "link_id": link_id,
            "table_row_id": row_id,
            "other_table_row_id": other_row_id,
        }

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Create Row Links (1:n)
    async def create_row_links(
        self, table_name: str, other_table_name: str, link_id: str, row_id: str, other_rows_ids: List[str]
    ):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/links/"

        other_rows_ids = other_rows_ids if isinstance(other_rows_ids, list) else [other_rows_ids]

        json = {
            "table_name": table_name,
            "other_table_name": other_table_name,
            "link_id": link_id,
            "row_id": row_id,
            "other_rows_ids": other_rows_ids,
        }

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Create Row Links Batch (m:n)
    async def create_row_links_batch(
        self,
        table_id: str,
        other_table_id: str,
        link_id: str,
        row_id_list: List[str],
        other_rows_ids_map: List[str],
    ):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-update-links/"

        json = {
            "table_id": table_id,
            "other_table_id": other_table_id,
            "link_id": link_id,
            "row_id_list": row_id_list,
            "other_rows_ids_map": other_rows_ids_map,
        }

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # List Row Links
    # [NOTE] Offset 처리해야 함 - limit 어디까지 가능한지 모름, default가 무한인지도 모름
    async def list_row_links(self, table_id: str, link_column: str, row_ids: List[str]):
        METHOD = "POST"
        URL = f"/dtable-db/api/v1/linked-records/{self.base_token.dtable_uuid}"

        json = {"table_id": table_id, "link_column": link_column, "rows": [{"row_id": row_id} for row_id in row_ids]}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    ################################################################
    # FILES & IMAGES
    ################################################################

    ################################################################
    # TABLES
    ################################################################
    # Add Table
    async def add_table(self, table_name: str):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {"table_name": table_name}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Create New Table
    # [NOTE] 이 endpoint는 Link 컬럼을 처리하지 못 함. (2023.9.9 현재)
    async def _create_new_table(self, table_name: str, columns: List[dict] = None):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {"table_name": table_name}
        if columns:
            json.update({"columns": columns})
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Rename Table
    async def rename_table(self, table_name: str, new_table_name: str):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {"table_name": table_name, "new_table_name": new_table_name}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Delete Table
    async def delete_table(self, table_name: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"
        ITEM = "success"

        json = {"table_name": table_name}
        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=json)
            results = response[ITEM]
        return results

    # Duplicate Table
    async def duplicate_table(self, table_name: str, is_duplicate_records: bool = True):
        # rename table in a second step
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/duplicate-table/"

        json = {"table_name": table_name, "is_duplicate_records": is_duplicate_records}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    ################################################################
    # IMPORT
    ################################################################
    # (DEPRECATED) Create Table from CSV

    # (DEPRECATED) Append Rows from CSV

    ################################################################
    # VIEWS
    ################################################################
    # List Views
    async def list_views(self, table_name: str, model: BaseModel = View, refresh: bool = True):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/"
        ITEM = "views"

        if not refresh and table_name in self.views:
            return self.views[table_name]

        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, table_name=table_name)
            results = response[ITEM]
        if model:
            results = [model(**x) for x in results]

        self.views.update({table_name: results})
        return self.views[table_name]

    # Create View
    async def create_view(
        self,
        table_name: str,
        name: str,
        type: str = "table",
        is_locked: bool = False,
        model: BaseModel = View,
    ):
        """
        type: "table" or "archive" (bigdata)
        """
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/"

        json = {"name": name, "type": type, "is_locked": str(is_locked).lower()}
        async with self.session_maker() as session:
            results = await self.request(
                session=session,
                method=METHOD,
                url=URL,
                json=json,
                table_name=table_name,
            )
        if model:
            results = model(**results)
        return results

    # Get View
    async def get_view(self, table_name: str, view_name: str, model: BaseModel = View):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/{view_name}/"

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, table_name=table_name)
        if model:
            results = model(**results)
        return results

    # Update View
    # NOT TESTED!
    async def update_view(self, table_name: str, view_name: str, conf: Union[dict, BaseModel] = None):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/{view_name}/"

        if isinstance(conf, BaseModel):
            conf = conf.dict()
        async with self.session_maker() as session:
            results = await self.request(
                session=session,
                method=METHOD,
                url=URL,
                json=conf,
                table_name=table_name,
            )
        return results

    # Delete View
    async def delete_view(self, table_name: str, view_name: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/{view_name}/"
        ITEM = "success"

        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, table_name=table_name)
            results = response[ITEM]
        return results

    ################################################################
    # COLUMNS
    ################################################################
    # Insert Column
    async def insert_column(
        self, table_name: str, column_name: str, column_type: str, column_data: dict = None, anchor_column: str = None
    ):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"

        json = await self.ensure_column(
            table_name=table_name,
            column_name=column_name,
            column_type=column_type,
            column_data=column_data,
            anchor_column=anchor_column,
        )
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Append Columns
    # [NOTE] 이 endpoint는 Link 컬럼을 처리하지 못 함. (2023.9.9 현재)
    async def append_columns(self, table_name: str, columns: List[dict]):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-append-columns/"

        columns = await asyncio.gather(*[self.ensure_column(table_name=table_name, **column) for column in columns])
        json = {"table_name": table_name, "columns": list()}
        for column in columns:
            column.pop("table_name")
            json["columns"].append(column)
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Update Column
    async def modify_column_type(self, table_name: str, column_name: str, new_column_type: str):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"

        json = {
            "op_type": "modify_column_type",
            "table_name": table_name,
            "column": column_name,
            "new_column_type": new_column_type,
        }
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Rename Column
    async def rename_column(self, table_name: str, column_name: str, new_column_name: str):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"

        json = {
            "op_type": "rename_column",
            "table_name": table_name,
            "column": column_name,
            "new_column_name": new_column_name,
        }
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Resize Column
    async def resize_column(self, table_name: str, column_name: str, new_column_width: int):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"

        json = {
            "op_type": "resize_column",
            "table_name": table_name,
            "column": column_name,
            "new_column_width": new_column_width,
        }
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Move Column
    async def move_column(self, table_name: str, column_name: str, target_column: str):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"

        json = {
            "op_type": "move_column",
            "table_name": table_name,
            "column": column_name,
            "target_column": target_column,
        }
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Freeze Column
    async def freeze_column(self, table_name: str, column_name: str, frozen: bool = True):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"

        json = {"op_type": "freeze_column", "table_name": table_name, "column": column_name, "frozen": frozen}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        return results

    # Delete Column
    async def delete_column(self, table_name: str, column_name: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/columns/"
        ITEM = "success"

        json = {"table_name": table_name, "column": column_name}
        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=json)
            results = response[ITEM]
        return results

    # Add Select Options
    async def add_select_options(
        self, table_name: str, column_name: str, options: List[SelectOption], model: BaseModel = None
    ):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/column-options/"

        json = {
            "table_name": table_name,
            "column": column_name,
            "options": [opt.dict(exclude_none=True) for opt in options],
        }
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)
        if model:
            results = model(**results)
        return results

    ################################################################
    # BIG DATA
    ################################################################

    ################################################################
    # ROW COMMENTS
    ################################################################
    async def list_row_comments(self, row_id: str):
        # NOT WORKING
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/comments/"

        params = {"row_id": row_id}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)
        return results

    ################################################################
    # NOTIFICATION
    ################################################################

    ################################################################
    # ACTIVITIES & LOGS
    ################################################################
    # Get Base Activity Logs
    async def get_base_activity_log(self, page: int = 1, per_page: int = 25, model: BaseModel = BaseActivity):
        # rename table in a second step
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/operations/"
        ITEM = "operations"

        params = {"page": page, "per_page": per_page}
        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, **params)
            results = response[ITEM]
        if model:
            results = [model(**x) for x in results]
        return results

    # List Row Activities
    async def list_row_activities(self, row_id: str, page: int = 1, per_page: int = 25):
        # rename table in a second step
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/activities/"

        params = {"row_id": row_id, "page": page, "per_page": per_page}
        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)
        return results

    # List Delete Operation Logs
    async def list_delete_operation_logs(self, op_type: str, page: int = 1, per_page: int = 25):
        """
        op_type: delete_row, delete_rows, delete_table, delete_column
        """
        # rename table in a second step
        METHOD = "GET"
        URL = f"/api/v2.1/dtables/{self.base_token.dtable_uuid}/delete-operation-logs/"
        ITEM = "delete_operation_logs"

        params = {"op_type": op_type, "page": page, "per_page": per_page}
        async with self.session_maker() as session:
            response = await self.request(session=session, method=METHOD, url=URL, **params)
            results = response[ITEM]
        return results

    # List Delete Rows
    async def list_delete_rows(self):
        # rename table in a second step
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/deleted-rows/"

        async with self.session_maker() as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        return results

    ################################################################
    # SNAPSHOTS
    ################################################################
    # TBD

    ################################################################
    # Helper
    ################################################################
    # Column Validator
    async def ensure_column(
        self, table_name: str, column_name: str, column_type: str, column_data: dict = None, anchor_column: str = None
    ):
        column = {"table_name": table_name, "column_name": column_name, "column_type": column_type}
        column_data = {} if not column_data else column_data

        # for single-select, mulitple-select - auto index number
        if column_type in ["single-select", "multiple-select"] and "options" in column_data:
            options = column_data["options"]
            if all([option.get("id") is None for option in options]):
                for id, option in enumerate(options, start=1):
                    option.update({"id": id})
                column_data.update({"options": options})

        # for link - auto add table
        if column_type in ["link"] and "table" not in column_data:
            column_data.update({"table": table_name})

        # for button
        if column_type in ["button"]:
            table = await self.get_table(table_name=table_name)
            table_id = table.id

            for button_action in column_data["button_action_list"]:
                # update current_table_id
                button_action.update({"current_table_id": table_id})

                # [NOTE] contact_email or name to email
                #  - input: {to_users: ["some_user@mail.com", "some_user_name"]}
                #  - output: {to_users: [{"value": "jlkajdfald@auth.local"}, {"value": "kajsldfasdf@auth.local"}]}
                if "to_users" in button_action:
                    collaborators = await self.list_collaborators()
                    to_users = button_action.pop("to_users")
                    to_user_ids = list()
                    for user in to_users:
                        for collaborator in collaborators:
                            if collaborator.name == user or collaborator.contact_email == user:
                                to_user_ids.append({"value": collaborator.email})
                    button_action.update({"to_users": to_user_ids})
                else:
                    button_action.update({"to_users": []})

                # [NOTE] user_column to user_col_key
                #  - input: {to_users: ["some_user@mail.com", "some_user_name"]}
                #  - output: {to_users: [{"value": "jlkajdfald@auth.local"}, {"value": "kajsldfasdf@auth.local"}]}
                if "user_column" in button_action:
                    user_column = button_action.pop("user_column")
                    for c in table.columns:
                        if c.name == user_column:
                            user_col_key = c.key
                            break
                    else:
                        _msg = f"user_column '{user_column}' is not in table '{table_name}'!"
                        raise KeyError(_msg)
                    button_action.update({"user_col_key": user_col_key})
            print(column_data)

        corrected_column_data = COLUMN_DATA[column_type](**column_data)
        column.update({"column_data": corrected_column_data.dict()})

        if anchor_column:
            column.update({"anchor_column": anchor_column})

        return column



import logging
from abc import abstractmethod
from datetime import datetime
from functools import partial
from typing import Any, Callable, Dict, List, Union

from ...model import Column, Table, User, Metadata
from ...utils import parse_str_datetime

logger = logging.getLogger(__name__)


SYSTEM_COLUMNS = [
    {"name": "_id", "key": "_id", "type": "row-id", "data": None},
    {"name": "_locked", "key": "_locked", "type": "checkbox", "data": None},
    {"name": "_locked_by", "key": "_locked_by", "type": "text", "data": None},
    {"name": "_archived", "key": "_archived", "type": "checkbox", "data": None},
    {"name": "_creator", "key": "_creator", "type": "creator", "data": None},
    {"name": "_ctime", "key": "_ctime", "type": "ctime", "data": None},
    {"name": "_mtime", "key": "_mtime", "type": "mtime", "data": None},
    {"name": "_last_modifier", "key": "_last_modifier", "type": "last-modifier", "data": None},
]

USER_FIELDS = ["user", "collaborator", "creator", "last-midifier"]


class CreateDeserializerFailed(Exception):
    pass


class DeserializeError(Exception):
    pass


class ColumnDeserializer:
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Column = None,
        collaborator_map: dict = None,
    ):
        self.name = name
        self.seatable_type = seatable_type
        self.data = data
        self.metadata = metadata
        self.collaborator_map = collaborator_map

    def __call__(self, x):
        if not x:
            return None
        try:
            return self.convert(x)
        except:
            _msg = f"deserialize failed: {self.seatable_type}({x})."
            raise DeserializeError(_msg)

    def schema(self):
        raise NotImplementedError

    def convert(self, x):
        raise NotImplementedError

    def get_table(self, table_name: str):
        for table in self.metadata.tables:
            if table.name == table_name:
                return table
        else:
            _msg = f"table '{table_name}' not exists!"
            raise KeyError(_msg)

    def get_table_by_id(self, table_id: str):
        for table in self.metadata.tables:
            if table.id == table_id:
                return table
        else:
            _msg = f"table id '{table_id}' not exists!"
            raise KeyError(_msg)

    def get_column_by_id(self, table_id: str, column_id: str):
        table = self.get_table_by_id(table_id=table_id)
        for column in table.columns:
            if column.key == column_id:
                return column
        else:
            _msg = f"no column (id: {column_id}) in table (id: {table_id})."
            raise KeyError(_msg)


class Deserializer:
    def __init__(
        self,
        metadata: Metadata,
        table_name: str,
        base_name: str = None,
        group_name: str = None,
        table_name_sep: str = "__",
        collaborators: List[User] = None,
    ):
        self.metadata = metadata
        self.table_name = table_name
        self.base_name = base_name
        self.group_name = group_name
        self.table_name_sep = table_name_sep
        self.collaborators = collaborators

        # get table
        for table in self.metadata.tables:
            if table.name == table_name:
                break
        else:
            _msg = f"table '{table_name}' not exists!"
            raise KeyError(_msg)
        self.table = table

        # get collaborator_map
        if self.collaborators:
            self.collaborator_map = {
                collaborator.email: f"{collaborator.name}({collaborator.contact_email})"
                for collaborator in self.collaborators
            }
        else:
            self.collaborator_map = None

        # prefix
        prefix = []
        if self.group_name:
            prefix.append(self.group_name)
        if self.base_name:
            prefix.append(self.base_name)
        self.table_name_prefix = self.table_name_sep.join(prefix)

        # helper
        self.mtime_column = None
        self.last_modified = None

        self.init_columns()

    @property
    @abstractmethod
    def Deserializer(self): ...

    @abstractmethod
    def schema(self): ...

    def generate_table_name(self):
        if self.table_name_prefix:
            return self.table_name_sep.join([self.table_name_prefix, self.table.name])
        return self.table.name

    def init_columns(self):
        LINK_REQUIRED = ["link", "link-formula"]
        COLLABORATOR_REQUIRED = ["user", "collaborator", "creator", "last-modifier"]

        column_keys = [c.key for c in self.table.columns]
        columns = [
            *[c.dict() for c in self.table.columns],
            *[c for c in SYSTEM_COLUMNS if c["key"] not in column_keys],
        ]

        self.columns = dict()
        for c in columns:
            # update column deserializer
            try:
                metadata = self.metadata if c["type"] in LINK_REQUIRED else None
                collaborator_map = self.collaborator_map if c["type"] in COLLABORATOR_REQUIRED else None
                deseriailizer = self.Deserializer[c["type"]](
                    name=c["name"],
                    seatable_type=c["type"],
                    data=c["data"],
                    metadata=metadata,
                    collaborator_map=collaborator_map,
                )
            except Exception:
                _msg = "create column deserializer failed - name: '{name}', seatable_type: '{type}', data: '{data}'.".format(
                    **c
                )
                raise CreateDeserializerFailed(_msg)
            self.columns.update({c["name"]: deseriailizer})

            # we need '_mtime' always!
            if c["type"] == "mtime" and c["name"] != "_mtime":
                self.mtime_column = c["name"]

        # we need '_mtime' always!
        if self.mtime_column:
            for c in SYSTEM_COLUMNS:
                if c["name"] == "_mtime":
                    self.columns.update(
                        {
                            "_mtime": self.Deserializer[c["type"]](
                                name=c["name"],
                                seatable_type=c["type"],
                                data=c["data"],
                            )
                        }
                    )
                    break

    def __call__(self, *row, select: list = None):
        if row is None:
            return

        if select == "*":
            select = None
        if select and not isinstance(select, list):
            select = [select]

        self.last_modified = None
        deserialized_rows = list()
        for r in row:
            deserialized_row = dict()
            for name in self.columns:
                if select and name not in select:
                    continue
                if name not in r:
                    continue
                value = self.columns[name](r[name])
                deserialized_row.update({name: value})
                if self.mtime_column and name == self.mtime_column:
                    self.last_modified = parse_str_datetime(r[name])
            if not select and self.mtime_column:
                deserialized_row.update({"_mtime": deserialized_row[self.mtime_column]})
            deserialized_rows.append(deserialized_row)

        return deserialized_rows



from enum import Enum

from pydantic import BaseModel, Field

from .core import ColumnType, _Model

__all__ = ["NewColumn", "SelectOption"]


class NumberFormat(Enum):
    number = "number"
    percent = "percent"
    dollar = "dollar"
    euro = "euro"
    yuan = "yuan"


class DateFormat(Enum):
    date = "YYYY-MM-DD"
    datetime = "YYYY-MM-DD HH:mm"


class DurationFormat(Enum):
    hm: str = "H:mm"
    hms: str = "H:mm:ss"


class RateType(Enum):
    rate = "dtable-icon-rate"
    like = "dtable-icon-like"
    praise = "dtable-icon-praise"
    flag = "dtable-icon-flag"


################################################################
# Data
################################################################
#
class ColumnData(_Model):
    pass


# Number Data
class NumberData(ColumnData):
    format: NumberFormat
    decimal: str = "dot"
    thousands: str = "comma"


# Date Data
class DateData(ColumnData):
    format: DateFormat = "datetime"


# Duration Data
class DurationData(ColumnData):
    format: str = "duration"
    duration_format: DurationFormat = "hms"


# Rating Data
class RatingData(ColumnData):
    rate_max_number: int = 5
    rate_style_color: str = "#EB00B1"
    rate_style_type: RateType = "dtable-icon-rate"


# Select Data (Single, Multi)
class SelectData(ColumnData):
    id: int
    name: str
    color: str
    text_color: str = Field(..., alias="text-color")


################################################################
# Data
################################################################
# Column
class NewColumn(_Model):
    column_name: str = Field(..., alias="name")
    column_type: ColumnType = Field(..., alias="type")
    column_data: ColumnData = Field(None, alias="data")
    anchor_column: str = None


# SelectOption
class SelectOption(_Model):
    name: str = None
    color: str = "gray"
    textColor: str = "white"



from datetime import datetime

from .const import DT_FMT, TZ


# divide chunks
def divide_chunks(x: list, chunk_size: int):
    for i in range(0, len(x), chunk_size):
        yield x[i : i + chunk_size]


# parse string datetime
def parse_str_datetime(x):
    if x.endswith("Z"):
        x = x.replace("Z", "+00:00", 1)
    try:
        x = datetime.strptime(x, DT_FMT)
    except Exception:
        x = datetime.fromisoformat(x)
    return x.astimezone(TZ)



################################
# UNDER DEV. - DO NOT USE
# ToPostgres에 맞추어서 다시 작성할 것!
################################

import logging

from plantable import model as pm
from plantable.const import DT_FMT, TZ

from ...model import Table
from .deserializer import Deserializer

logger = logging.getLogger(__name__)


AVRO_BOOLEAN_NULLABLE = ["null", "int"]
AVRO_STRING = "string"
AVRO_STRING_NULLABLE = ["null", "string"]
AVRO_INT_NULLABLE = ["null", "int"]
AVRO_LONG_NULLABLE = ["null", "long"]
AVRO_DOUBLE_NULLABLE = ["null", "double"]
AVRO_DATE_NULLABLE = ["nulll", {"type": "int", "logicalType": "date"}]
AVRO_TIMESTAMP = {"type": "int", "logicalType": "timestamp-millis"}
AVRO_TIMESTAMP_NULLABLE = ["nulll", {"type": "int", "logicalType": "timestamp-millis"}]
AVRO_DURATION_NULLABLE = ["nulll", {"type": "int", "logicalType": "duration"}]
AVRO_STRING_ARRAY_NULLABLE = ["null", {"type": "array", "items": "string"}]


def convert_number(column: pm.Column):
    if column.data and column.data.get("enable_precision") and column.data["precision"] == 0:
        return {"name": column.name, "type": AVRO_LONG_NULLABLE}
    return {"name": column.name, "type": AVRO_DOUBLE_NULLABLE}


def convert_date(column):
    if column.data and column.data["format"] == "YYYY-MM-DD":
        return {"name": column.name, "type": AVRO_DATE_NULLABLE}
    return {"name": column.name, "type": AVRO_TIMESTAMP_NULLABLE}


SCHEMA_MAP = {
    "checkbox": lambda column: {"name": column.name, "type": AVRO_BOOLEAN_NULLABLE},
    "text": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "string": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "button": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "long-text": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "email": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "url": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "rate": lambda column: {"name": column.name, "type": AVRO_INT_NULLABLE},
    "number": convert_number,
    "date": convert_date,
    "duration": lambda column: {"name": column.name, "type": AVRO_DURATION_NULLABLE},
    "ctime": lambda column: {"name": column.name, "type": AVRO_TIMESTAMP},
    "mtime": lambda column: {"name": column.name, "type": AVRO_TIMESTAMP},
    "single-select": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "multiple-select": lambda column: {"name": column.name, "type": AVRO_STRING_ARRAY_NULLABLE},
    "link": lambda column: {"name": column.name, "type": AVRO_STRING_ARRAY_NULLABLE},  # CHECK
    "link-formula": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "user": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "collaborator": lambda column: {"name": column.name, "type": AVRO_STRING_ARRAY_NULLABLE},  # CHECK
    "creator": lambda column: {"name": column.name, "type": AVRO_STRING},
    "last-modifier": lambda column: {"name": column.name, "type": AVRO_STRING},
    "file": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "image": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "formula": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
    "auto-number": lambda column: {"name": column.name, "type": AVRO_STRING_NULLABLE},
}

SYSTEM_COLUMNS = {
    "_locked": {"name": "_locked", "type": AVRO_BOOLEAN_NULLABLE},
    "_locked_by": {"name": "_locked_by", "type": AVRO_STRING_NULLABLE},
    "_archived": {"name": "_archived", "type": AVRO_BOOLEAN_NULLABLE},
    "_ctime": {"name": "_ctime", "type": AVRO_TIMESTAMP},
    "_mtime": {"name": "_mtime", "type": AVRO_TIMESTAMP},
    "_creator": {"name": "_creator", "type": AVRO_STRING},
    "_last_modifier": {"name": "_last_modifier", "type": AVRO_STRING},
}


################################################################
# Converter
################################################################
class ToAvro(Deserializer):
    def init_schema(self):
        # copy system columns
        hidden_fields = SYSTEM_COLUMNS.copy()

        # add fields
        fields = [{"name": "_id", "type": AVRO_STRING}]
        for c in self.table.columns:
            fields.append(SCHEMA_MAP[c.type](c))
            if c.key in hidden_fields:
                hidden_fields.pop(c.key)

        # ramained hidden fields
        for c in hidden_fields:
            fields.append(hidden_fields[c])

        return {"namespace": "", "type": "record", "name": self.table.name, "fields": fields}



import logging

import click
import uvicorn
from click_loglevel import LogLevel

logger = logging.getLogger(__file__)


@click.group()
def plantable():
    pass


@plantable.command()
def hello():
    print("Hello, Plantable!")


@plantable.group()
def server():
    pass


@server.command()
@click.option("-h", "--host", type=str, default="0.0.0.0")
@click.option("-p", "--port", type=int, default=3000)
@click.option("--reload", is_flag=True)
@click.option("--workers", type=int, default=None)
@click.option("--log-level", type=LogLevel(), default=logging.INFO)
def run(host, port, reload, workers, log_level):
    logging.basicConfig(level=log_level)

    if reload:
        app = "plantable.server.app:app"
    else:
        from .server.app import app

    uvicorn.run(
        app,
        host=host,
        port=port,
        reload=reload,
        workers=workers,
        log_level=log_level,
    )
